# OpenTrafficShaper webserver module: users page
# Copyright (C) 2007-2013, AllWorldIT
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

package opentrafficshaper::plugins::webserver::pages::statistics;

use strict;
use warnings;


# Exporter stuff
require Exporter;
our (@ISA,@EXPORT,@EXPORT_OK);
@ISA = qw(Exporter);
@EXPORT = qw(
);
@EXPORT_OK = qw(
);


use DateTime;
use HTML::Entities;
use HTTP::Status qw( :constants );
use JSON;

use opentrafficshaper::logger;
use opentrafficshaper::plugins;
use opentrafficshaper::utils qw( parseURIQuery );

use opentrafficshaper::plugins::configmanager qw( getLimit getInterface isTrafficClassValid getTrafficClassName );

use opentrafficshaper::plugins::statistics::statistics;



# Graphs by limit
sub bylimit
{
	my ($kernel,$globals,$client_session_id,$request) = @_;


	# Header
	my $content = <<EOF;
		<div id="header">
			<h2>Limit Stats View</h2>
		</div>
EOF

	my $limit;

	# Check request
	if ($request->method eq "GET") {
		# Parse GET data
		my $queryParams = parseURIQuery($request);
		# We need our LID
		if (!defined($queryParams->{'lid'})) {
			$content .=<<EOF;
				<tr class="info">
					<td colspan="8"><p class="text-center">No LID in Query String</p></td>
				</tr>
EOF
			goto END;
		}
		# Check if we get some data back when pulling the limit from the backend
		if (!defined($limit = getLimit($queryParams->{'lid'}))) {
			$content .=<<EOF;
				<tr class="info">
					<td colspan="8"><p class="text-center">No Results</p></td>
				</tr>
EOF
			goto END;
		}
	}

	my $name = (defined($limit->{'FriendlyName'}) && $limit->{'FriendlyName'} ne "") ? $limit->{'FriendlyName'} : $limit->{'Username'};
	my $usernameEncoded = encode_entities($name);


	# Build content
	$content = <<EOF;
	<div id="content" style="float:left">

		<div style="position: relative; top: 50px;">
			<h4 style="color:#8f8f8f;">Latest Data For: $usernameEncoded</h4>
		<br/>

		<div id="ajaxData" class="ajaxData" style="float:left; width:1024px; height: 560px"></div>
		</div>

	</div>
EOF

	#$content .= statistics::do_test();
#	$content .= opentrafficshaper::plugins::statistics::do_test();

	# Files loaded at end of HTML document
	my @javascripts = (
		'/static/awit-flot/jquery.flot.min.js',
		'/static/awit-flot/jquery.flot.time.min.js',
#		'/static/awit-flot/jquery.flot.websockets.js'
	);

	# Build our data path using the URI module to make sure its nice and clean
	my $dataPath = URI->new('/statistics/data-by-limit');
	# Pass it the original query, just incase we had extra params we can use
	$dataPath->query_form( $request->uri->query_form() );
	my $dataPathStr = $dataPath->as_string();


	# String put in <script> </script> tags after the above files are loaded
	my $javascript = _getJavascript($dataPathStr);

END:

	return (HTTP_OK,$content,{ 'javascripts' => \@javascripts, 'javascript' => $javascript });
}



# Return data by limit
sub databylimit
{
	my ($kernel,$globals,$client_session_id,$request) = @_;


	# Parse GET data
	my $queryParams = parseURIQuery($request);

	# Check if the limit ID was passed to us
	if (!defined($queryParams->{'lid'})) {
		return (HTTP_OK,{ 'error' => 'Invalid lid' },{ 'type' => 'json' });
	}

	my $limit;
	if (!defined($limit = getLimit($queryParams->{'lid'}))) {
		return (HTTP_OK,{ 'error' => 'Invalid limit' },{ 'type' => 'json' });
	}

	# Pull in stats data
	my $statsData = opentrafficshaper::plugins::statistics::getStatsByLID($queryParams->{'lid'});

	# First stage refinement
	my $rawData;
	foreach my $timestamp (sort keys %{$statsData}) {
		foreach my $direction (keys %{$statsData->{$timestamp}}) {

			foreach my $stat ('rate','pps','cir','limit') {
				push(  @{$rawData->{"$direction.$stat"}->{'data'}} , [ $timestamp , $statsData->{$timestamp}->{$direction}->{$stat} ] );
			}

		}
	}
	# Second stage - add labels
	foreach my $direction ('tx','rx') {
		foreach my $stat ('rate','pps','cir','limit') {
				# Make it looks nice:  Tx Rate
				my $label = uc($direction) . " " . ucfirst($stat);
				# And set it as the label
				$rawData->{"$direction.$stat"}->{'label'} = $label;
		}
	}

	# Final stage, chop it out how we need it
	my $jsonData = [ $rawData->{'tx.limit'}, $rawData->{'rx.limit'}, $rawData->{'tx.rate'} , $rawData->{'rx.rate'} ];

	return (HTTP_OK,$jsonData,{ 'type' => 'json' });
}


# Graphs by class
sub byclass
{
	my ($kernel,$globals,$client_session_id,$request) = @_;


	# Header
	my $content = <<EOF;
		<div id="header">
			<h2>Class Stats View</h2>
		</div>
EOF

	my $interface;
	my $cid;

	# Check if its a GET request...
	if ($request->method eq "GET") {
		# Parse GET data
		my $queryParams = parseURIQuery($request);
		# Grab the interface name
		if (!defined($queryParams->{'interface'})) {
			$content .=<<EOF;
				<tr class="info">
					<td colspan="8"><p class="text-center">No interface in Query String</p></td>
				</tr>
EOF
			goto END;
		}
		# Grab the class
		if (!defined($queryParams->{'class'})) {
			$content .=<<EOF;
				<tr class="info">
					<td colspan="8"><p class="text-center">No class in Query String</p></td>
				</tr>
EOF
			goto END;
		}
		# Check if we get some data back when pulling the interface from the backend
		if (!defined($interface = getInterface($queryParams->{'interface'}))) {
			$content .=<<EOF;
				<tr class="info">
					<td colspan="8"><p class="text-center">No Interface Results</p></td>
				</tr>
EOF
			goto END;
		}
		# Check if our traffic class is valid
		if (!defined($cid = isTrafficClassValid($queryParams->{'class'})) && $queryParams->{'class'} ne "0") {
			$content .=<<EOF;
				<tr class="info">
					<td colspan="8"><p class="text-center">No Class Results</p></td>
				</tr>
EOF
			goto END;
		}
	}

	my $interfaceNameEncoded = encode_entities($interface->{'name'});

	my $classNameEncoded;
	if ($cid) {
		$classNameEncoded = encode_entities(getTrafficClassName($cid));
	} else {
		$classNameEncoded = $interfaceNameEncoded;
	}


	# Build content
	$content = <<EOF;
	<div id="content" style="float:left">

		<div style="position: relative; top: 50px;">
			<h4 style="color:#8f8f8f;">Latest Data For: $classNameEncoded on $interfaceNameEncoded</h4>
		<br/>

		<div id="ajaxData" class="ajaxData" style="float:left; width:1024px; height: 560px"></div>
		</div>

	</div>
EOF

	# Files loaded at end of HTML document
	my @javascripts = (
		'/static/awit-flot/jquery.flot.min.js',
		'/static/awit-flot/jquery.flot.time.min.js',
	);

	# Build our data path using the URI module to make sure its nice and clean
	my $dataPath = URI->new('/statistics/data-by-class');
	# Pass it the original query, just incase we had extra params we can use
	$dataPath->query_form( $request->uri->query_form() );
	my $dataPathStr = $dataPath->as_string();


	# String put in <script> </script> tags after the above files are loaded
	my $javascript = _getJavascript($dataPathStr);

END:

	return (HTTP_OK,$content,{ 'javascripts' => \@javascripts, 'javascript' => $javascript });
}



# Return data by class
sub databyclass
{
	my ($kernel,$globals,$client_session_id,$request) = @_;


	# Parse GET data
	my $queryParams = parseURIQuery($request);

	# Check if the username was passed to us
	if (!defined($queryParams->{'interface'})) {
		return (HTTP_OK,{ 'error' => 'Missing interface' },{ 'type' => 'json' });
	}
	if (!defined($queryParams->{'class'})) {
		return (HTTP_OK,{ 'error' => 'Missing class' },{ 'type' => 'json' });
	}
	if (!defined(my $iface = getInterface($queryParams->{'interface'}))) {
		return (HTTP_OK,{ 'error' => 'Invalid interface' },{ 'type' => 'json' });
	}
	if (!defined(my $cid = isTrafficClassValid($queryParams->{'class'})) && $queryParams->{'class'} ne "0") {
		return (HTTP_OK,{ 'error' => 'Invalid class' },{ 'type' => 'json' });
	}

	# Pull in stats data
	my $statsData = opentrafficshaper::plugins::statistics::getStatsByClass($queryParams->{'interface'},$queryParams->{'class'});

	# First stage refinement
	my $rawData;
	foreach my $timestamp (sort keys %{$statsData}) {
		foreach my $direction (keys %{$statsData->{$timestamp}}) {

			foreach my $stat ('rate','pps','cir','limit') {
				push(  @{$rawData->{"$direction.$stat"}->{'data'}} , [ $timestamp , $statsData->{$timestamp}->{$direction}->{$stat} ] );
			}

		}
	}

	my $jsonData = [ ];

	# Second stage - add labels
	foreach my $direction ('tx','rx') {
		foreach my $stat ('rate','pps','cir','limit') {
				# Make it looks nice:  Tx Rate
				my $label = uc($direction) . " " . ucfirst($stat);
				# And set it as the label
				$rawData->{"$direction.$stat"}->{'label'} = $label;
		}

		# JSON stuff we looking for
		foreach my $stat ('limit','rate') {
			if (defined($rawData->{"$direction.$stat"}->{'data'})) {
					push(@{$jsonData},$rawData->{"$direction.$stat"});
			}
		}
	}


	return (HTTP_OK,$jsonData,{ 'type' => 'json' });
}


# Return javascript for the graph
sub _getJavascript
{
	my $dataPath = shift;


	my $javascript =<<EOF;
// Tooltip - Displays detailed information regarding the data point
function showTooltip(x, y, contents) {
	jQuery('<div id="tooltip">' + contents + '</div>').css({
			position: 'absolute',
			display: 'none',
			top: y - 30,
			left: x - 50,
			color: "#fff",
			padding: '2px 5px',
			'border-radius': '6px',
			'background-color': '#000',
			opacity: 0.80
	}).appendTo("body").fadeIn(200);
}

var previousPoint = null;
jQuery("#ajaxData").bind("plothover", function (event, pos, item) {
	if (item) {
		if (previousPoint != item.dataIndex) {
			previousPoint = item.dataIndex;
			jQuery("#tooltip").remove();
			showTooltip(item.pageX, item.pageY, item.series.label);
		}
	} else {
		jQuery("#tooltip").remove();
		previousPoint = null;
	}
});


// Setting up the graph here
options = {

	series: {
		lines: {
			show: true,
			lineWidth: 1,
			fill: true,
			fillColor: {
				colors: [
					{ opacity: 0.1 },
					{ opacity: 0.13 }
				]
			}
		},

		points: {
			show: false,
			lineWidth: 2,
			radius: 3
		},

		shadowSize: 0,
		stack: true
	},

	grid: {
		hoverable: true,
		clickable: false,
		tickColor: "#f9f9f9",
		borderWidth: 0
	},

	legend: {
		labelBoxBorderColor: "#fff"
	},

	xaxis: {
		mode: "time",

		tickSize: [60, "second"],

		tickFormatter: function (v, axis) {
			var date = new Date(v);

			if (date.getSeconds() % 5 == 0) {
				var hours = date.getHours() < 10 ? "0" + date.getHours() : date.getHours();
				var minutes = date.getMinutes() < 10 ? "0" + date.getMinutes() : date.getMinutes();
				var seconds = date.getSeconds() < 10 ? "0" + date.getSeconds() : date.getSeconds();

				return hours + ":" + minutes + ":" + seconds;
			} else {
				return "";
			}
		},
	},

	yaxis: {
		min: 0,
//		max: 4000,
		tickFormatter: function (v, axis) {
			if (v % 10 == 0) {
				res = v.toString().replace(/\\B(?=(\\d{3})+(?!\\d))/g, ",");
				console.log(res);
				return res + ' Kbps';
			} else {
				return "";
			}
		},
	}
}

// Load data from ajax
jQuery.ajax({
	url: '$dataPath',
	dataType: 'json',

	success: function(statsData) {
		plot = null;

		// formatting time to match javascript's epoch in milliseconds
		for (i = 0; (i < statsData.length); i++) {
			//console.log(statsData[i].data);
			for (y = 0; (y < statsData[i].data.length); y++) {
				d = new Date(statsData[i].data[y][0] * 1000);
				statsData[i].data[y][0] = statsData[i].data[y][0] * 1000;
			}
		}

		if (statsData.length > 0) {
			plot = jQuery.plot(jQuery("#ajaxData"), statsData, options);
		}
	}
});
EOF

	return $javascript;
}


1;
